@@ -1,9 +1,12 @@
-#ifndef NVIM_CMDHIST_H
-#define NVIM_CMDHIST_H
+#pragma once
 
-#include "nvim/eval/typval.h"
-#include "nvim/ex_cmds_defs.h"
-#include "nvim/os/time.h"
+#include <stddef.h>  // IWYU pragma: keep
+
+#include "nvim/cmdexpand_defs.h"  // IWYU pragma: keep
+#include "nvim/eval/typval_defs.h"  // IWYU pragma: keep
+#include "nvim/ex_cmds_defs.h"  // IWYU pragma: keep
+#include "nvim/os/time_defs.h"
+#include "nvim/types_defs.h"  // IWYU pragma: keep
 
 /// Present history tables
 typedef enum {
@@ -16,18 +19,17 @@ typedef enum {
   HIST_DEBUG,         ///< Debug commands.
 } HistoryType;
 
-/// Number of history tables
-#define HIST_COUNT      (HIST_DEBUG + 1)
+enum { HIST_COUNT = HIST_DEBUG + 1, };  ///< Number of history tables
 
 /// History entry definition
-typedef struct hist_entry {
+typedef struct {
   int hisnum;           ///< Entry identifier number.
   char *hisstr;         ///< Actual entry, separator char after the NUL.
+  size_t hisstrlen;     ///< Length of hisstr (excluding the NUL).
   Timestamp timestamp;  ///< Time when entry was added.
-  list_T *additional_elements;  ///< Additional entries from ShaDa file.
+  AdditionalData *additional_data;  ///< Additional entries from ShaDa file.
 } histentry_T;
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "cmdhist.h.generated.h"
 #endif
-#endif  // NVIM_CMDHIST_H
